package net.lapisblue.server.images;

import java.awt.*;
import java.awt.image.*;
import java.text.*;
import java.io.*;
import java.util.*;
import javax.imageio.*;
import javax.servlet.*;
import javax.servlet.http.*;


public abstract class ImageServlet extends HttpServlet {
  protected final Properties defaults = new Properties();
  protected final LinkedList cache = new LinkedList();
  static final int maxCacheSize = 400;
  final String cachePath = "/image-cache";
  final String serviceName;
  File cacheDir;
  File cacheMap;
  
  static {
    System.setProperty("java.awt.headless", "true");
  }
  
  protected ImageServlet () {
    final String fullname = this.getClass().getName();
    final String classname = fullname.substring(fullname.lastIndexOf('.') + 1);
    
    if (classname.endsWith("Servlet")) {
      serviceName = classname.substring(0, classname.length() - 7).toLowerCase();
    } else {
      serviceName = classname.toLowerCase();
    }
   
    try {
      defaults.load(getClass().getResourceAsStream(
        serviceName + ".properties"));
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }
  
  class CacheEntry {
    final String query;
    final File file;
    
    CacheEntry (String line) {
      final int at = line.indexOf(" => ");
      query = line.substring(at + 4); 
      file = new File(cacheDir, line.substring(0, at));
    }
    
    CacheEntry (String query, File file) {
      this.query = query;
      this.file = file;
    }
    
    File file () { return file; }
    String path () { return cachePath + "/" + file.getName(); }
    
    public String toString () { return file.getName() + " => " + query; }
  }
  
  public final void init () {
    try {
      cacheDir = new File(getServletContext().getRealPath(cachePath));
      
      if (!cacheDir.exists()) {
        if (!cacheDir.mkdirs()) {
          throw new RuntimeException("cannot create " + cacheDir.getPath());
        }
      }
      
      cacheMap = new File(cacheDir, serviceName + ".map");
      
      if (cacheMap.exists()) {
        final BufferedReader reader = new BufferedReader(new FileReader(cacheMap));
      
        String line;

        while ((line = reader.readLine()) != null) {
          final CacheEntry entry = new CacheEntry(line);
          
          if (entry.file().exists()) {
            cache.add(entry);
          }
        }
      }
      
      final File[] files = cacheDir.listFiles();
      
      if (files != null) {
        checkingFiles:
        for (int index = 0, length = files.length; index < length; index++) {
          final File file = files[index];
        
          if (file.getName().startsWith(serviceName) &&
              file.getName().endsWith(".png")) {
            for (Iterator cacheIt = cache.iterator(); cacheIt.hasNext(); ) {
              if (((CacheEntry)cacheIt.next()).file().equals(file)) {
                continue checkingFiles;
              }
            }
        
            file.delete();
          }
        }
      }
    } catch (IOException ioe) {
      getServletContext().log("cache init", ioe);
    }
    
    subclassInit() ;
  }
  
  public void subclassInit () { }

  public final void destroy () {
    writeCacheMap();
  }
  
  private void writeCacheMap () {
    synchronized (cache) {
      try {
        PrintStream out = null;
        try {
          out = new PrintStream(new FileOutputStream(cacheMap));
      
          for (Iterator cacheIt = cache.iterator(); cacheIt.hasNext(); ) {
            out.println(cacheIt.next());
          }
        } finally {
          if (out != null) { out.close(); }
        }
      } catch (IOException ioe) {
        getServletContext().log("cache write failed", ioe);
      }
    }
  }
  
  protected abstract class ImageGenerator {
    final HttpServletRequest request;
    final Color background;
    
    protected ImageGenerator (HttpServletRequest request) {
      this.request = request;
      this.background = getColor("bg");
    }

    protected abstract int getImageWidth () ;
    protected abstract int getImageHeight () ;

    protected final void paint (Graphics2D g2D) {
      if (background != null) {
        g2D.setPaint(background);
        g2D.fillRect(0,0,getImageWidth(),getImageHeight());
      }
      
      draw(g2D);
    }
    
    protected int getInt (String name) {
      String param = getString(name);

      try {
        return Integer.parseInt(param);
      } catch (NumberFormatException nfe) {
      } catch (NullPointerException npe) {
      }

      throw new IllegalArgumentException(name + "=" + param);
    }
    
    protected double getDouble (String name) {
      String param = getString(name);

      try {
        return Double.parseDouble(param);
      } catch (NumberFormatException nfe) {
      } catch (NullPointerException npe) {
      }

      throw new IllegalArgumentException(name + "=" + param);
    }
    
    protected float getFloat (String name) {
      String param = getString(name);

      try {
        return Float.parseFloat(param);
      } catch (NumberFormatException nfe) {
      } catch (NullPointerException npe) {
      }

      throw new IllegalArgumentException(name + "=" + param);
    }
    
    protected Color getColor (String name) {
      final String param = getString(name);
      
      if (param == null) { return null; }
      
      try {
        final int red;
        final int green;
        final int blue;
        final int alpha;

        if (param.length() <= 4) {
          red = 0x11 * Character.digit(param.charAt(0), 16);
          green = 0x11 * Character.digit(param.charAt(1), 16);
          blue = 0x11 * Character.digit(param.charAt(2), 16);
          alpha = param.length() == 4 ? (0x11 * Character.digit(param.charAt(3), 16)) : 0xff;
        } else {
          red = Integer.parseInt(param.substring(0,2), 16);
          green = Integer.parseInt(param.substring(2,4), 16);
          blue = Integer.parseInt(param.substring(4,6), 16);
          alpha = param.length() == 8 ? Integer.parseInt(param.substring(7,8), 16) : 0xff;
        }
        
        return new Color(red, green, blue, alpha);
      } catch (NumberFormatException nfe) {
      } catch (NullPointerException npe) {
      }

      throw new IllegalArgumentException(name + "=" + param);
    }
    
    protected String getString (String name) {
      final String param = (request != null) ? request.getParameter(name) : null;
      
      return (param == null) ? defaults.getProperty(name) : param;
    }
    
    protected BufferedImage generateImage () {
      final BufferedImage image = new BufferedImage(
        getImageWidth(), getImageHeight(), BufferedImage.TYPE_INT_ARGB );
    
      final Graphics2D g2d = (Graphics2D)image.getGraphics();

      paint(g2d);
    
      g2d.dispose();

      return image;
    }
    
    protected abstract void draw (Graphics2D g2d) ;
  }
  
  protected abstract ImageGenerator createImageGenerator (HttpServletRequest request) ;
    
  public void doGet (HttpServletRequest request, HttpServletResponse response)
    throws IOException, ServletException {
    final String referrer = request.getHeader("referrer");
    
    if ((request.getPathInfo() != null) && 
        (request.getPathInfo().endsWith(".css") ||
         request.getPathInfo().endsWith(".xsl"))) {
      response.sendRedirect(contextPath(request) + "/style" + request.getPathInfo());
    }
    
    final String query = request.getQueryString();
    
    /*
    if ((referrer != null) &&
        (!referrer.startsWith("http://" + request.getServerName()))) {
      final PrintWriter out = new PrintWriter(new OutputStreamWriter(
        response.getOutputStream(), "UTF-8"));
      response.setContentType("text/xml; charset=UTF-8");
      out.println("<denied "+
                  "referrer='" + referrer + "' " + 
                  "server='" + request.getServerName() + "'/>");
      out.flush();
      
      return;
    }
    */  
    
    if ((query == null) || request.getParameter("parameters") != null) {
      response.setContentType("text/xml; charset=UTF-8");

      final InputStream in = getClass().getResourceAsStream(serviceName + "-parameters.xml");
      final OutputStream out = response.getOutputStream();
      int octet;
      
      while ((octet = in.read()) != -1) { out.write((byte)octet); }

      return;
    }
    
    if (request.getParameter("image-types") != null) {
      final PrintWriter out = new PrintWriter(new OutputStreamWriter(
        response.getOutputStream(), "UTF-8"));
      response.setContentType("text/xml; charset=UTF-8");
      out.println("<types>");
      
      final String[] mimeTypes = ImageIO.getWriterMIMETypes();
      final String[] names = ImageIO.getWriterFormatNames();
      
      for (int index = 0, length = names.length; index < length; index++) {
        out.println("<name>" + names[index] + "</name>");
      }

      for (int index = 0, length = mimeTypes.length; index < length; index++) {
        out.println("<type>" + mimeTypes[index] + "</type>");
      }
      
      out.println("</types>");
      out.flush();
      
      return;
    }
    
    if (request.getParameter("list-cache") != null) {
      final PrintWriter out = new PrintWriter(new OutputStreamWriter(
        response.getOutputStream(), "UTF-8"));
      response.setContentType("text/xml; charset=UTF-8");
      out.println("<cache size='" + cache.size() + "'>");
      
      final String contextPath = contextPath(request);
      
      synchronized (cache) {
        for (Iterator cacheIt = cache.iterator(); cacheIt.hasNext(); ) {
          final CacheEntry entry = (CacheEntry)cacheIt.next();
        
          out.println("  <entry "+
                      "href='" + contextPath + entry.path() + "' " + 
                      "query='" + entry.query.replaceAll("\\'", "&apos;").replaceAll("\\&", "&amp;") + "'" +
                      "/>");
        }
      }
      
      out.println("</cache>");
      out.flush();
      
      return;
    }
    
    if (request.getParameter("flush-cache") != null) {
      final PrintWriter out = new PrintWriter(new OutputStreamWriter(
        response.getOutputStream(), "UTF-8"));
      response.setContentType("text/xml; charset=UTF-8");
      
      synchronized (cache) {
        for (Iterator cacheIt = cache.iterator(); cacheIt.hasNext(); cacheIt.remove()) {
          ((CacheEntry)cacheIt.next()).file().delete();
        }
      }      
      out.println("<cache/>");
      out.flush();
      
      return;
    }

    synchronized (cache) {
      for (Iterator cacheIt = cache.iterator(); cacheIt.hasNext(); ) {
        final CacheEntry entry = (CacheEntry)cacheIt.next();
      
        if (query.equals(entry.query)) {
          response.sendRedirect(contextPath(request) + entry.path());
          
          return;
        }
      }
    
      while (cache.size() >= maxCacheSize) {
        final CacheEntry last = (CacheEntry)cache.getLast();
        cache.removeLast();
        last.file().delete();
      }
    }
    
    final ImageGenerator generator = createImageGenerator(request);
    
    final BufferedImage image = generator.generateImage();
    
    response.setContentType("image/png");
    
    final File file = File.createTempFile(serviceName, ".png", cacheDir);
    final CacheEntry entry = new CacheEntry(query, file);
    final ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        
    ImageIO.write( image, "png", buffer);
    
    buffer.writeTo(response.getOutputStream());
    
    FileOutputStream out = null;
    
    try {
      out = new FileOutputStream(file);
      buffer.writeTo(out);
    
      synchronized(cache) {
        cache.add(entry);
        
        writeCacheMap();
      }
    } finally {
      if (out != null) {
        out.close();
      }
    }
    
    response.flushBuffer();
  }
  
  private String contextPath (HttpServletRequest request) {
    final String contextPath = request.getContextPath();
          
    return (contextPath.endsWith("/") ?
            contextPath.substring(0, contextPath.length() - 1) :
            contextPath);
  }
}